#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Matplotlib-based animation classes.
'''

#FIXME: To avoid memory leaks, I'm fairly certain that everywhere we currently
#call the ".lines.remove()" method of a Matplotlib streamplot object, that we
#instead need to simply call remove(): e.g.,
#
#    # ...instead of this:
#    self._stream_plot.lines.remove()
#
#    # ...do this:
#    self._stream_plot.remove()
#
#The latter removes the stream plot itself from the parent figure and axes, in
#preparation for recreating the stream plot. The former merely removes the set
#of all stream lines from the stream plot. Failing to remove the stream plot
#itself could potentially cause old stream plots to be retained in memory. In
#any case, give self._stream_plot.remove() a go; if that works, we'll want to
#use that regardless of whether there are any memory leaks here or not.
#FIXME: Sadly, the self._stream_plot.remove() method has yet to be fully
#implemented (...at least, under older Matplotlib versions). Is there no better
#approach than replacing the entire streamplot each animation frame? We're
#fairly certain that we've seen Matplotlib example code that updates an
#existing streamplot each animation frame instead. Investigate the aged pandas!

# ....................{ IMPORTS                            }....................
import numpy as np
from betse.lib.numpy import nparray
from betse.science.math import mathunit
from betse.science.visual.anim.animafter import (
    AnimCellsAfterSolving, AnimVelocity)
from betse.science.visual.plot.plotutil import cell_mosaic, cell_mesh
from betse.util.type.types import type_check, SequenceTypes
from matplotlib.collections import LineCollection, PolyCollection
from scipy import interpolate

# ....................{ CLASSES ~ after                    }....................
#FIXME: This class should probably no longer be used, now that the Gouraud
#shading performed by the "AnimCellsMembranesData" class has been optimized.
class AnimFlatCellsTimeSeries(AnimCellsAfterSolving):
    '''
    Animation of an arbitrary cell-centric time series (e.g., average cell
    voltage as a function of time).

    Attributes
    ----------
    _cell_plot : ???
        Artists signifying cell data for the prior or current frame.
    _cell_time_series : list
        Arbitrary cell data as a function of time to be underlayed.
    _gapjunc_plot : LineCollection
        Lines signifying gap junction state for the prior or current frame.
    _gapjunc_time_series : list
        Arbitrary gap junction data as a function of time to be overlayed.
    '''

    @type_check
    def __init__(
        self,
        time_series: SequenceTypes,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this animation.

        Parameters
        ----------
        cell_time_series : SequenceTypes
            Arbitrary cell data as a function of time

        See the superclass `__init__()` method for all remaining parameters.
        '''

        # Initialize the superclass.
        super().__init__(*args, time_step_count=len(time_series), **kwargs)

        # Classify all remaining parameters.
        self._cell_time_series = time_series

        # Cell data series for the first frame.
        data_set = self._cell_time_series[0]

        # Add a collection of cell polygons with animated voltage data.
        if self._phase.p.showCells is True:
            self._cell_plot, self._axes = cell_mosaic(
                data_set, self._axes, self._phase.cells, self._phase.p, self._colormap)
        else:
            self._cell_plot, self._axes = cell_mesh(
                data_set, self._axes, self._phase.cells, self._phase.p, self._colormap)

        # Display and/or save this animation.
        self._animate(
            color_mappables=self._cell_plot,
            color_data=self._cell_time_series,
        )


    @type_check
    def _plot_frame_figure(self):

        # Cell data series for this frame.
        zv = self._cell_time_series[self._time_step]
        if self._phase.p.showCells is True:
            zz_grid = zv
        else:
            zz_grid = zv
            # zz_grid = np.zeros(len(self._phase.cells.voronoi_centres))
            # zz_grid[self._phase.cells.cell_to_grid] = zv

        # Update the cell plot for this frame.
        self._cell_plot.set_array(zz_grid)


class AnimEnvTimeSeries(AnimCellsAfterSolving):
    '''
    Animation of an arbitrary cell-agnostic time series (e.g., environmental
    voltage as a function of time), plotted over the cell cluster.

    Attributes
    ----------
    _time_series : list
        Arbitrary environmental data as a function of time to be plotted.
    _mesh_plot : matplotlib.image.AxesImage
        Meshplot of the current or prior frame's environmental data.
    '''

    @type_check
    def __init__(
        self,
        time_series: SequenceTypes,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this animation.

        Parameters
        ----------
        time_series : SequenceTypes
            Arbitrary environmental data as a function of time to be plotted.

        See the superclass `__init__()` method for all remaining parameters.
        '''

        # Initialize the superclass.
        super().__init__(*args, **kwargs)

        # Classify parameters required by the _plot_frame_figure() method.
        self._time_series = time_series

        # Environmental data meshplot for the first frame.
        self._mesh_plot = self._plot_image(pixel_data=self._time_series[0])

        # Display and/or save this animation.
        self._animate(
            color_mappables=self._mesh_plot,
            color_data=self._time_series,
        )


    def _plot_frame_figure(self) -> None:

        # Environmental data meshplot for this frame.
        self._mesh_plot.set_data(self._time_series[self._time_step])


#FIXME: Disable display of the cell-centric time series, leaving only the
#gap junction-centric time series displayed. Alternately, might it be possible
#to filter the cell-centric time series with some sort of alpha-based fading
#effect -- reducing the prominance of that series but not entirely eliminating
#that series? Contemplate us up, anyways.

class AnimGapJuncTimeSeries(AnimCellsAfterSolving):
    '''
    Animation of an arbitrary gap junction-centric time series (e.g., the gap
    junction open state as a function of time) overlayed an arbitrary cell-
    centric time series (e.g., cell voltage as a function of time) on the cell
    cluster.

    Attributes
    ----------
    _cell_plot : ???
        Artists signifying cell data for the prior or current frame.
    _cell_time_series : list
        Arbitrary cell data as a function of time to be underlayed.
    _gapjunc_plot : LineCollection
        Lines signifying gap junction state for the prior or current frame.
    _gapjunc_time_series : list
        Arbitrary gap junction data as a function of time to be overlayed.
    '''

    @type_check
    def __init__(
        self,
        time_series: SequenceTypes,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this animation.

        Parameters
        ----------
        cell_time_series : SequenceTypes
            Arbitrary cell data as a function of time to be underlayed.
        gapjunc_time_series : SequenceTypes
            Arbitrary gap junction data as a function of time to be overlayed.

        See the superclass `__init__()` method for all remaining parameters.
        '''

        # Initialize the superclass.
        super().__init__(*args, time_step_count=len(time_series), **kwargs)

        # Classify all remaining parameters.
        # self._cell_time_series = cell_time_series
        self._time_series = time_series

        # Gap junction data series for the first frame plotted as lines.
        self._gapjunc_plot = LineCollection(
            mathunit.upscale_coordinates(self._phase.cells.nn_edges),
            array=self._time_series[0],
            cmap=self._phase.p.gj_cm,
            linewidths=2.0,
            zorder=10,
        )
        # self._gapjunc_plot.set_clim(0.0, 1.0)
        self._axes.add_collection(self._gapjunc_plot)

        # Cell data series for the first frame.
        # data_set = self._cell_time_series[0]

        # # Add a collection of cell polygons with animated voltage data.
        # if self.p.showCells is True:
        #     self._cell_plot, self._axes = cell_mosaic(
        #         data_set, self._axes, self.cells, self.p, self._colormap)
        # else:
        #     self._cell_plot, self._axes = cell_mesh(
        #         data_set, self._axes, self.cells, self.p, self._colormap)

        # Display and/or save this animation.
        self._animate(
            color_mappables=self._gapjunc_plot,
            color_data=self._time_series,
        )

    #
    # @type_check
    def _plot_frame_figure(self):

        # Update the gap junction plot for this frame.
        self._gapjunc_plot.set_array(
            self._time_series[self._time_step])

        # # Cell data series for this frame.
        # zv = self._cell_time_series[self._time_step]
        # if self.p.showCells is True:
        #     zz_grid = zv
        # else:
        #     zz_grid = np.zeros(len(self.cells.voronoi_centres))
        #     zz_grid[self.cells.cell_to_grid] = zv

        # Update the cell plot for this frame.
        # self._cell_plot.set_array(zz_grid)


class AnimMembraneTimeSeries(AnimCellsAfterSolving):
    '''
    Animation of an arbitrary cell membrane-specific time series (e.g.,
    membrane channel or pump density factor as a function of time), plotted
    over the cell cluster.

    This factor changes in response to changes in electroosmotic and
    electrophoretic movements, produced by self-generated fields and flows in
    the cell cluster.

    Attributes
    ----------
    _mem_edges : LineCollection
        Membrane edges coloured for the current or prior frame.
    _time_series : SequenceTypes
        Arbitrary cell membrane data as a function of time to be plotted.
    '''


    @type_check
    def __init__(
        self,
        time_series: SequenceTypes,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this animation.

        Parameters
        ----------
        time_series : Sequence
            Arbitrary cell membrane data as a function of time to be plotted.

        See the superclass `__init__()` method for all remaining parameters.
        '''

        # Initialize the superclass.
        super().__init__(*args, time_step_count=len(time_series), **kwargs)

        # Classify parameters required by the _plot_frame_figure() method.
        self._time_series = time_series

        # Membrane edges coloured for the first frame.
        self._mem_edges = LineCollection(
            self._phase.cells.mem_edges_flat * self._phase.p.um,
            array=self._time_series[0],
            cmap=self._colormap,
            linewidths=4.0,
        )
        self._axes.add_collection(self._mem_edges)

        # Display and/or save this animation.
        self._animate(
            color_mappables=self._mem_edges,
            color_data=self._time_series,
        )


    def _plot_frame_figure(self) -> None:

        # Update membrane edges colours for this frame.
        self._mem_edges.set_array(
            self._time_series[self._time_step])


#FIXME: This animation class no longer appears to be used. Consider excising.
#FIXME: Before we eliminate this class entirely, it would be quite useful to
#ensure that the "PolyCollection" logic performed below exists as a
#"LayerCellsABC" subclass.
class AnimMorphogenTimeSeries(AnimCellsAfterSolving):
    '''
    Animation of the concentration of an arbitrary morphogen in both cells and
    the environment as a function of time, plotted over the cell cluster.

    Parameters
    ----------
    _cell_time_series : Sequence
        Morphogen concentration in cells as a function of time.
    _env_time_series : Sequence
        Morphogen concentration in the environment as a function of time.
    '''


    @type_check
    def __init__(
        self,
        cell_time_series: SequenceTypes,
        env_time_series: SequenceTypes,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this animation.

        Parameters
        ----------
        cell_time_series : Sequence
            Morphogen concentration in cells as a function of time.
        env_time_series : Sequence
            Morphogen concentration in the environment as a function of time.

        See the superclass `__init__()` method for all remaining parameters.
        '''

        # Initialize the superclass.
        super().__init__(*args, time_step_count=len(cell_time_series), **kwargs)

        # Classify the passed parameters.
        self._cell_time_series = cell_time_series
        self._env_time_series = env_time_series

        #FIXME: Rename:
        #
        #* "bkgPlot" to "_"... we have no idea. Animate first. Decide later.
        #* "collection" to "_mesh_plot".

        self.bkgPlot = self._plot_image(
            pixel_data=self._env_time_series[0].reshape(self._phase.cells.X.shape))

        #FIXME: Try reducing to: self.cells.cell_verts * self.p.um

        # Polygon collection based on individual cell polygons.
        points = np.multiply(self._phase.cells.cell_verts, self._phase.p.um)
        self.collection = PolyCollection(
            points, cmap=self._colormap, edgecolors='none')
        self.collection.set_array(self._cell_time_series[0])
        self._axes.add_collection(self.collection)

        # Display and/or save this animation.
        self._animate(
            color_mappables=(self.collection, self.bkgPlot),

            # If colorbar autoscaling is requested, clip the colorbar to the
            # minimum and maximum morphogen concentrations -- regardless of
            # whether that morphogen resides in cells or the environment.
            color_data=(
                self._cell_time_series, self._env_time_series),
        )


    def _plot_frame_figure(self) -> None:

        self.collection.set_array(
            self._cell_time_series[self._time_step])
        self.bkgPlot.set_data(
            self._env_time_series[self._time_step].reshape(
                self._phase.cells.X.shape))

# ....................{ SUBCLASSES ~ velocity              }....................
class AnimVelocityIntracellular(AnimVelocity):
    '''
    Animation of fluid velocity over all intracellular gap junctions plotted on
    the cell cluster.

    Attributes
    -----------
    _mesh_plot : matplotlib.image.AxesImage
        Meshplot of the current or prior frame's velocity field magnitude.
    _stream_plot : matplotlib.streamplot.StreamplotSet
        Streamplot of the current or prior frame's velocity field _or_ `None`
        if such field has yet to be streamplotted.
    '''


    def __init__(self, *args, **kwargs) -> None:

        # Initialize the superclass.
        super().__init__(*args, **kwargs)

        # Define this attribute *BEFORE* streamplotting, which assumes this
        # attribute to exist.
        self._stream_plot = None

        #FIXME: Inefficient. This streamplot will be recreated for the first
        #time step in the exact same manner; so, it's unclear that we need to
        #do so here.

        # Streamplot the first frame's velocity field.
        vfield, vnorm = self._plot_stream_velocity_field(time_step=0)

        # Meshplot the first frame's velocity field magnitude.
        self._mesh_plot = self._plot_image(
            pixel_data=vfield,
            colormap=self._phase.p.background_cm,
        )

        #FIXME: How expensive would caching these calculations be? Oh, just do
        #it already. We have to calculate these values *ANYWAY*, so there's no
        #incentive at all in delaying the matter.

        # Display and/or save this animation. Since recalculating "vfield" for
        # each animation frame is non-trivial, this call avoids passing the
        # "time_series" parameter. Instead, the _get_velocity_field() method
        # manually rescales the colorbar on each frame according to the minimum
        # and maximum velocity field magnitude. While non-ideal, every
        # alternative is currently worse.
        self._animate(color_mappables=self._mesh_plot)


    def _plot_frame_figure(self) -> None:

        # Streamplot this frame's velocity field.
        vfield, vnorm = self._plot_stream_velocity_field(self._time_step)

        # Meshplot this frame's velocity field.
        self._mesh_plot.set_data(vfield)

        # Rescale the colorbar if needed.
        self._mesh_plot.set_clim(self._color_min, self._color_max)


    @type_check
    def _plot_stream_velocity_field(self, time_step: int) -> tuple:
        '''
        Streamplot the current velocity field for the passed frame and return a
        2-tuple describing this field.

        Returns
        ----------
        (Sequence, float)
            2-element tuple `(velocity_field, velocity_field_magnitude_max)`
            whose:
            * First element is the list of all velocity field magnitudes for
              this frame.
            * Second element is the maximum such magnitude.
        '''

        cell_centres = (
            self._phase.cells.cell_centres[:, 0],
            self._phase.cells.cell_centres[:, 1])
        cell_grid = (self._phase.cells.X, self._phase.cells.Y)

        #FIXME: Ugh. Duplicate code already performed by the superclass
        #AnimCellsABC._init_current_density() method. We clearly need a
        #general-purpose interpolation utility method. Hawkish doves in a cove!
        u_gj_x = self._phase.cells.maskECM * interpolate.griddata(
            cell_centres,
            self._phase.sim.u_cells_x_time[time_step],
            cell_grid,
            fill_value=0,
            method=self._phase.p.interp_type,
        )
        u_gj_y = self._phase.cells.maskECM * interpolate.griddata(
            cell_centres,
            self._phase.sim.u_cells_y_time[time_step],
            cell_grid,
            fill_value=0,
            method=self._phase.p.interp_type,
        )

        # Current velocity field magnitudes and the maximum such magnitude.
        vfield = np.sqrt(u_gj_x**2 + u_gj_y**2) * 1e9
        vnorm = np.max(vfield)

        # Streamplot the current velocity field for this frame.
        self._stream_plot = self._plot_stream(
            old_stream_plot=self._stream_plot,
            x=u_gj_x / vnorm,
            y=u_gj_y / vnorm,
            magnitude=vfield,
            magnitude_max=vnorm,
        )
        # self.streamV.set_UVC(u_gj_x/vnorm,u_gj_y/vnorm)

        # Rescale the colorbar range if desired.
        if self._conf.is_color_autoscaled:
            self._color_min = np.min(vfield)
            self._color_max = vnorm

        return (vfield, vnorm)


class AnimVelocityExtracellular(AnimVelocity):
    '''
    Animation of fluid velocity over all extracellular spaces plotted on the
    cell cluster.

    Attributes
    ----------
    _mesh_plot : matplotlib.image.AxesImage
        Meshplot of the current or prior frame's velocity field magnitude.
    _stream_plot : matplotlib.streamplot.StreamplotSet
        Streamplot of the current or prior frame's velocity field.
    _magnitude_time_series : np.ndarray
        Time series of all fluid velocity magnitudes.
    '''

    def __init__(self, *args, **kwargs) -> None:

        # Initialize the superclass.
        super().__init__(*args, is_ecm_required=True, **kwargs)

        # Time series of all velocity magnitudes.
        self._magnitude_time_series = np.sqrt(
            nparray.from_iterable(self._phase.sim.u_env_x_time) ** 2 +
            nparray.from_iterable(self._phase.sim.u_env_y_time) ** 2) * 1e6

        # Velocity field and maximum velocity field value for the first frame.
        vfield = self._magnitude_time_series[0]
        vnorm = np.max(vfield)

        # Velocity field meshplot for the first frame.
        self._mesh_plot = self._plot_image(
            pixel_data=vfield,
            colormap=self._phase.p.background_cm,
        )

        #FIXME: Doesn't this streamplot the last frame instead?

        # Velocity field streamplot for the first frame.
        self._stream_plot = self._axes.quiver(
            self._phase.cells.xypts[:, 0] * self._phase.p.um,
            self._phase.cells.xypts[:, 1] * self._phase.p.um,
            self._phase.sim.u_env_x_time[-1].ravel() / vnorm,
            self._phase.sim.u_env_y_time[-1].ravel() / vnorm,
        )

        # Display and/or save this animation.
        self._animate(
            color_mappables=self._mesh_plot,
            color_data=self._magnitude_time_series,
        )


    def _plot_frame_figure(self) -> None:

        # Velocity field and maximum velocity field value for this frame.
        vfield = self._magnitude_time_series[self._time_step]
        vnorm = np.max(vfield)

        # Update the current velocity meshplot.
        self._mesh_plot.set_data(vfield)

        # Update the current velocity streamplot.
        self._stream_plot.set_UVC(
            self._phase.sim.u_env_x_time[self._time_step] / vnorm,
            self._phase.sim.u_env_y_time[self._time_step] / vnorm,
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level matplotlib-specific functionality for which no better home exists.
'''

# ....................{ IMPORTS                            }....................
from betse.util.io.error.errwarning import ignoring_warnings
from beartype.typing import ContextManager
from matplotlib import MatplotlibDeprecationWarning

# ....................{ WARNINGS                           }....................
def ignoring_deprecations_mpl() -> ContextManager:
    '''
    Single-shot context manager temporarily ignoring all matplotlib-specific
    deprecation warnings emitted by the :mod:`warnings` module for the duration
    of this context.

    See Also
    -----------
    :class:`ignoring_warnings`
        Further details.
    '''

    return ignoring_warnings(MatplotlibDeprecationWarning)



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Voltage-gated channel classes based on the Morris-Lecar simplified Hodgkin-Huxley channel approximations.
'''

# .................... IMPORTS                            ....................
from abc import ABCMeta, abstractmethod
import numpy as np
from betse.science.channels.channelsabc import ChannelsABC
from betse.util.io.log import logs

# .................... BASE                               ....................
class VgKABC(ChannelsABC, metaclass=ABCMeta):
    '''
    Abstract base class of all Voltage-gated Morris-Lecar channel classes.

    Attributes
    ----------
    _mInf :
        Equillibrium value function for m-gates of channel
    _mTau :
        Time-constant function for m-gates of channel

    '''

    def init(self, vm, cells, p, targets = None):
        '''
        Initialize targeted voltage-gated sodium channels at the initial
        time step of the simulation based on the initial cell Vmems.

        Channel model uses Hodgkin-Huxley kinetic model style
        for voltage gated channels.
        '''

        if targets is None:

            self.targets = None

            if type(vm) == float or type(vm) == np.float64:
                self.data_length = 1

            else:
                self.data_length = len(vm)


        else:
            self.targets = targets
            self.data_length = len(self.targets)
            self.mdl = len(cells.mem_i)

        self.modulator = 1.0

        if self.targets is None:

            V = vm * 1000

        else:
            V = vm[self.targets] * 1000

        self._init_state(V)

    def run(self, vm, p):
        '''
        Handle all targeted voltage-gated sodium channels by working with the passed
        user-specified parameters on the tissue simulation and cellular
        world for a time step.

        Channel model uses Hodgkin-Huxley kinetic model style
        for voltage gated channels.

        '''

        if self.targets is None:

            V = vm*1000

        else:

            V = vm[self.targets] * 1000

        self._calculate_state(V)

        if self.kinetic_gate: # If the gate dynamics are time-dependent:

            self.update_ml(p, time_unit = self.time_unit)

        else: # the probability is simply the steady-state voltage-sensitivity:

            self.m = self._mInf

        if self.targets is None:

            self.P = self.m

        else:
            self.P = np.zeros(self.mdl)
            self.P[self.targets] = self.m


    @abstractmethod
    def _init_state(self, V):
        '''
        Initializes values of the m and h gates of the channel.
        '''
        pass

    @abstractmethod
    def _calculate_state(self, V):
        '''
        Calculates time-dependent values of the m and h gates of the channel.
        '''
        pass

# ....................{ SUBCLASS                           }....................

class Kv_ML1(VgKABC):
    '''
    Default delayed-rectifier voltage gated K+ channel model described in the Morris-Lecar literature.

    '''
    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the KV channel: Kv_MorrisLecar1 ')

        self.time_unit = 1.0e3

        self.vrev = -85     # reversal voltage used in model [mV]

        self.ions = ['K']
        self.rel_perm = [1]
        self.kinetic_gate = True
        self.Phi = 0.066  # time-constant

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 0.5*(1 + np.tanh((V - 12.0)/17.0))

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 0.5*(1 + np.tanh((V- 12)/17))
        self._mTau = 1/(np.cosh((V - 12)/(2*17)))

class Kv2p1_ML(VgKABC):
    '''
    Default delayed-rectifier voltage gated K+ channel model paralleling that of Korngreen et al, but using
    simplified Morris-Lecar formalism.

    '''
    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the KV channel: Kv2.1, MorrisLecar ')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        self.ions = ['K']
        self.rel_perm = [1]
        self.kinetic_gate = True
        self.Phi = 0.066

        # initialize values of the m gate of the channel based on m_inf:
        self.m = 0.5*(1 + np.tanh((V - 14.0)/30.0))

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 0.5*(1 + np.tanh((V- 14)/30))
        self._mTau = 1/(np.cosh((V - 14)/(2*30)))

class Kv1p3_ML(VgKABC):
    '''
    Default delayed-rectifier voltage gated K+ channel model paralleling the Kv1.3 channel of  et al, but using
    simplified Morris-Lecar formalism.

    '''
    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the KV channel: Kv1.3, Morris-Leclar ')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        self.ions = ['K']
        self.rel_perm = [1]
        self.kinetic_gate = True
        self.Phi = 0.066

        # initialize values of the m gate of the channel based on m_inf:
        self.m = 0.5*(1 + np.tanh((V + 14.0)/20.0))

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 0.5*(1 + np.tanh((V + 14)/20))
        self._mTau = 1/(np.cosh((V + 14)/(2*20)))

class Kv1p5_ML(VgKABC):
    '''
    Default delayed-rectifier voltage gated K+ channel model paralleling the Kv1.5 channel of  et al, but using
    simplified Morris-Lecar formalism.

    '''
    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the KV channel: Kv_1.5, Morris-Lecar ')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        self.ions = ['K']
        self.rel_perm = [1]
        self.kinetic_gate = True
        self.Phi = 0.066

        # initialize values of the m gate of the channel based on m_inf:
        self.m = 0.5*(1 + np.tanh((V + 6.0)/15.0))

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 0.5*(1 + np.tanh((V + 6)/15))
        self._mTau = 1/(np.cosh((V + 6)/(2*15)))

class Kv1p5S_ML(VgKABC):
    '''
    Slow version of delayed-rectifier voltage gated K+ channel model paralleling the Kv1.5 channel of  et al, but using
    simplified Morris-Lecar formalism, with extended time constant for large time-step simulations.

    '''
    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the KV channel: Kv_1.5, Morris-Lecar ')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        self.ions = ['K']
        self.rel_perm = [1]
        self.kinetic_gate = True
        self.Phi = 0.00066*2
        # self.Phi = 0.00066

        # initialize values of the m gate of the channel based on m_inf:
        self.m = 0.5*(1 + np.tanh((V + 6.0)/15.0))

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 0.5*(1 + np.tanh((V + 6)/15))
        self._mTau = 1/(np.cosh((V + 6)/(2*15)))

class Nav_ML(VgKABC):
    '''
    Persistent (i.e. inactivating) NaV channel current based on NaV1.6.

    '''
    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the NaV channel: Nav_MorrisLecar ')

        self.time_unit = 1.0e3

        self.vrev = 50     # reversal voltage used in model [mV]

        self.ions = ['Na']
        self.rel_perm = [1]
        self.kinetic_gate = False
        self.Phi = 1.0

        # initialize values of the m gate of the channel based on m_inf:
        self.m = 0.5*(1 + np.tanh((V + 17.0)/18.0))

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 0.5*(1 + np.tanh((V + 17)/18))
        self._mTau = 1

class Cav_L_ML(VgKABC):
    '''
    Persistent (i.e. inactivating) L-type CaV channel current based on model of Carlin et al (a CaV1.3 channel).

    '''
    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the CaV channel: Cav_L_MorrisLecar ')

        self.time_unit = 1.0e3

        self.vrev = 80     # reversal voltage used in model [mV]

        self.ions = ['Ca']
        self.rel_perm = [1]
        self.kinetic_gate = False
        self.Phi = 1.0

        # initialize values of the m gate of the channel based on m_inf:
        self.m = 0.5*(1 + np.tanh((V + 20.0)/24.0))

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 0.5*(1 + np.tanh((V + 20)/24))
        self._mTau = 1

class Cav_L_ML2(VgKABC):
    '''
    Persistent (i.e. inactivating) L-type CaV channel current based on model of Carlin et al (a CaV1.3 channel).

    '''
    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the CaV channel: Cav_L_MorrisLecar ')

        self.time_unit = 1.0e3

        self.vrev = 80     # reversal voltage used in model [mV]

        self.ions = ['Ca']
        self.rel_perm = [1]
        self.kinetic_gate = False
        self.Phi = 1.0

        # initialize values of the m gate of the channel based on m_inf:
        self.m = 0.5*(1 + np.tanh((V + 20.0)/12.0))

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 0.5*(1 + np.tanh((V + 20)/12))
        self._mTau = 1

class Cav_N_ML(VgKABC):
    '''
    Persistent (i.e. inactivating) N-type CaV channel current based on CaV2.2 model of Huang et al.

    '''
    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the CaV channel: Cav_N_MorrisLecar ')

        self.time_unit = 1.0e3

        self.vrev = 80     # reversal voltage used in model [mV]

        self.ions = ['Ca']
        self.rel_perm = [1]
        self.kinetic_gate = False
        self.Phi = 1.0

        # initialize values of the m gate of the channel based on m_inf:
        self.m = 0.5*(1 + np.tanh((V + 1.0)/18.0))

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 0.5*(1 + np.tanh((V + 1.0)/18.0))
        self._mTau = 1

class Cav_T_ML(VgKABC):
    '''
    Persistent (i.e. inactivating) T-type CaV channel current based on CaV3.1 model of Traboulsie et al.

    '''
    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the CaV channel: Cav_T_MorrisLecar ')

        self.time_unit = 1.0e3

        self.vrev = 80     # reversal voltage used in model [mV]

        self.ions = ['Ca']
        self.rel_perm = [1]
        self.kinetic_gate = False
        self.Phi = 1.0

        # initialize values of the m gate of the channel based on m_inf:
        self.m = 0.5*(1 + np.tanh((V + 43.0)/24.0))

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 0.5*(1 + np.tanh((V + 43)/24.0))
        self._mTau = 1

class Kir_ML(VgKABC):
    '''
    Inward-rectifier Kir voltage-sensitive K+ channel current based on Kir2.1 model of Samy et al.

    '''
    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the KV channel: Kir_MorrisLecar ')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        self.ions = ['K']
        self.rel_perm = [1]
        self.kinetic_gate = False
        self.Phi = 1.0

        # initialize values of the m gate of the channel based on m_inf:
        self.m = 1/(np.cosh((V + 135)/(37)))

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1/(np.cosh((V + 135)/(37)))
        self._mTau = 1


class HCN2_ML(VgKABC):
    '''
    HCN2 cation channel current based on HCN2 model of Moosmang et al.

    '''
    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the channel: HCN2_MorrisLecar ')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        self.ions = ['K', 'Na']
        self.rel_perm = [1, 0.2]
        self.kinetic_gate = False
        self.Phi = 1.0

        # initialize values of the m gate of the channel based on m_inf:
        self.m = 0.5*(1 + np.tanh((V + 99.0)/12.4))  #19.2

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 0.5*(1 + np.tanh((V + 99.0)/12.4))
        self._mTau = 1

class HCN4_ML(VgKABC):
    '''
    HCN4 cation channel current based on HCN2 model of Moosmang et al.

    '''
    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the channel: HCN2_MorrisLecar ')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        self.ions = ['K', 'Na']
        self.rel_perm = [1, 0.2]
        self.kinetic_gate = False
        self.Phi = 1.0

        # initialize values of the m gate of the channel based on m_inf:
        self.m = 0.5*(1 + np.tanh((V + 99.0)/19.2))  #19.2

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 0.5*(1 + np.tanh((V + 99.0)/19.2))
        self._mTau = 1



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all matplotlib-based layer subclasses spatially
plotting vectors onto the cell cluster.
'''

# ....................{ IMPORTS                            }....................
from betse.science.math.vector.veccls import VectorCellsCache
from betse.science.phase.phasecls import SimPhase
from betse.science.visual.layer.lyrabc import LayerCellsColorfulABC
from betse.util.type.types import type_check

# ....................{ SUBCLASSES                         }....................
class LayerCellsVectorColorfulABC(LayerCellsColorfulABC):
    '''
    Abstract base class of all classes spatially plotting vectors of arbitrary
    one-dimensional data (e.g., transmembrane voltage) onto the cell cluster
    such that the elements of these vectors are mappable as colors onto the
    colorbars of parent visuals.

    Attributes
    ----------
    _vector : VectorCellsCache
        Cache of various vectors of the same underlying data situated along
        different coordinate systems for all time steps to be animated.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, vector: VectorCellsCache) -> None:
        '''
        Initialize this layer.

        Parameters
        ----------
        vector : VectorCellsCache
            Cache of various vectors of the same underlying data situated along
            different coordinate systems for all time steps to be animated.
        '''

        # Initialize our superclass.
        super().__init__()

        # Classify all passed parameters.
        self._vector = vector

# ....................{ MAKERS                             }....................
#FIXME: Rename to layer_vector().
@type_check
def make_layer(phase: SimPhase, *args, **kwargs) -> LayerCellsVectorColorfulABC:
    '''
    Vector Layer suitable for spatially plotting the passed vector of arbitrary
    one-dimensional data (e.g., transmembrane voltage) for the passed simulation
    phase onto the cell cluster such that the elements of this vector are
    mappable as colors onto the colorbars of parent visuals.

    The type of layer instantiated and returned by this factory function
    conditionally depends on the configuration for this phase. (See this
    function's implementation for details.)

    Parameters
    ----------
    phase : SimPhase
        Current simulation phase.

    All remaining arguments are passed as is to the ``__init__`` method of this
    layer's class.

    Returns
    ----------
    LayerCellsVectorColorfulABC
        Layer suitable for plotting the passed vector.
    '''

    # Avoid circular import dependencies.
    from betse.science.visual.layer.vector.lyrvecdiscrete import (
        LayerCellsVectorDiscreteMembranesFixed,
        LayerCellsVectorDiscreteMembranesDeformed,
    )
    from betse.science.visual.layer.vector.lyrvecsmooth import (
        LayerCellsVectorSmoothRegions)

    # Type of layer to be created.
    layer_type = None

    # If this simulation configuration requests that each cell in the cell
    # cluster be plotted...
    if phase.p.showCells:
        # If this phase enables deformations, dynamically plot these cells
        # in a manner redrawing cell boundaries each time step.
        if phase.p.deformation:
            layer_type = LayerCellsVectorDiscreteMembranesDeformed
        # Else, this phase does *NOT* enable deformations. In this case,
        # statically plot these cells in a manner reusing the cell
        # boundaries plotted for the first time step on subsequent time
        # steps.
        else:
            layer_type = LayerCellsVectorDiscreteMembranesFixed
    # Else, this simulation configuration requests that the cell cluster be
    # plotted as a smooth continuum. In this case, do so.
    else:
        layer_type = LayerCellsVectorSmoothRegions

    # Create and return this layer.
    return layer_type(*args, **kwargs)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **Python-based multithreading** (i.e., multithreading implemented at
the Python API level rather than by third-party C extensions, libraries, or
packages) facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetsePyException
from betse.util.type.decorator.decmemo import func_cached

# ....................{ EXCEPTIONS                        }....................
def die_unless_gil() -> None:
    '''
    Raise an exception unless the active Python interpreter prohibits genuine
    multithreading via a Global Interpreter Lock (GIL).

    Equivalently, this function raises an exception if this interpreter is
    **GIL-less** (i.e., permits genuine multithreading and hence is *not*
    encumbered by a GIL).

    Raises
    ----------
    BetsePyException
        If this interpreter is GIL-less.

    See Also
    ----------
    :func:`is_gil`
        Further details.
    '''

    # If this Python interpreter is GIL-less, raise an exception.
    if not is_gil():
        raise BetsePyException(
            'Python interpreter GIL-less '
            '(i.e., permits genuine multithreading).')


def die_unless_threadable() -> None:
    '''
    Raise an exception unless the active Python interpreter supports
    multithreading (in principal) at the API level.

    Raises
    ----------
    BetsePyException
        If this interpreter does *not* support multithreading.

    See Also
    ----------
    :func:`is_threadable`
        Further details.
    '''

    # If this Python interpreter is *NOT* multithreadable, raise an exception.
    if not is_threadable():
        raise BetsePyException(
            'Python interpreter not multithreadable '
            '(i.e., "_thread" module not importable).')

# ....................{ TESTERS                           }....................
@func_cached
def is_gil() -> bool:
    '''
    ``True`` only if the active Python interpreter prohibits genuine
    multithreading via a Global Interpreter Lock (GIL).

    Caveats
    ----------
    This function returning ``True`` does *not* imply this interpreter to
    prohibit multithreading (in principal) at the API level. Interpreters that
    prohibit genuine multithreading via a GIL typically still at least support
    standard multithreading APIs, albeit in a manner silently reducing multi-
    to single-core processing. Consider calling the :func:`is_threadable`
    function to clarify this edge case.

    This function returns ``True`` only if this interpreter is either CPython
    (i.e., the reference Python implementation) *or* PyPy; all other
    interpreters (e.g., IronPython, Jython) are well-known to *not* prohibit
    genuine multithreading via a GIL. While an admittedly non-future-proofed
    hack, the Python API currently provides no portable means of querying for
    the "GIL-ness" of an interpreter.
    '''

    # Avoid circular import dependencies.
    from betse.util.py import pyimpl

    # Return true only if this is either CPython or PyPy. See above.
    return pyimpl.is_cpython() or pyimpl.is_pypy()


@func_cached
def is_threadable() -> bool:
    '''
    ``True`` only if the active Python interpreter supports multithreading (in
    principal) at the API level.

    Equivalently, this function returns ``True`` only if the low-level standard
    :mod:`_thread` module is importable under this interpreter. This module is
    strictly optional and hence available only for systems both supporting
    multithreading *and* compiling Python with such support.

    Systems supporting multithreading includes:

    * Windows.
    * macOS.
    * Linux.
    * SGI IRIX.
    * Solaris 2.x.
    * Systems implementing the POSIX-compliant ``pthread`` API.

    Caveats
    ----------
    This function returning ``True`` does *not* imply this interpreter to be
    **GIL-less** (i.e., permit genuine multithreading and hence is *not*
    encumbered by a GIL). Interpreters that prohibit genuine multithreading via
    a GIL typically still at least support standard multithreading APIs, albeit
    in a manner silently reducing multi- to single-core processing. Consider
    calling the :func:`is_gil` function to clarify this edge case.
    '''

    # Avoid circular import dependencies.
    from betse.util.py.module import pymodname

    # Return true only if a low-level optional module installed *ONLY* when
    # Python supports multithreading is importable.
    return pymodname.is_module('_thread')



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level standard output facilities.
'''

#FIXME: Copy setup.util.get_command_output() here after completion.

# ....................{ IMPORTS                           }....................

# ....................{ OUTPUTTERS                        }....................
def output(*objs) -> None:
    '''
    Print all passed objects as is to standard output *without* logging these
    objects.

    This function is provided only for orthogonality with the function of the
    same name defined by the :mod:`betse.util.io.stderrs` module.

    This function is intentionally *not* named ``print()`` to avoid conflict
    with the builtin function of the same name.
    '''

    print(*objs)


def output_lines(*objs) -> None:
    '''
    Print all passed objects as is to standard output *without* logging these
    objects, delimiting each such object by a newline.
    '''

    print('\n'.join(objs))


def output_traceback() -> None:
    '''
    Print the current call stack to standard output.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.call import callers

    # Print this call stack, excluding the calls to both this and the
    # callers.get_traceback() functions.
    output(callers.get_traceback(-2))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Top-level application namespace.

For PEP 8 compliance, this namespace exposes a subset of the metadata constants
provided by the :mod:`betse.metadata` module commonly inspected by external
automation.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during setuptools-based installation, this
# module may import *ONLY* from modules guaranteed to exist at the start of
# installation. This includes all standard Python and application modules but
# *NOT* third-party dependencies, which if currently uninstalled will only be
# installed at some later time in the installation. Likewise, to avoid circular
# import dependencies, the top-level of this module should avoid importing
# application modules where feasible.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ IMPORTS                            }....................
# Subject all subsequent imports to @beartype-based hybrid runtime-static
# type-checking *BEFORE* importing anything further.
# from beartype import BeartypeConf
from beartype.claw import beartype_this_package
beartype_this_package()

# For PEP 8 compliance, versions constants expected by external automation are
# imported under their PEP 8-mandated names.
from betse.metadata import VERSION as __version__
from betse.metadata import VERSION_PARTS as __version_info__

# ....................{ GLOBALS                            }....................
# Document all global variables imported into this namespace above.

__version__
'''
Human-readable application version as a ``.``-delimited string.

For PEP 8 compliance, this specifier has the canonical name ``__version__``
rather than that of a typical global (e.g., ``VERSION_STR``).
'''


__version_info__
'''
Machine-readable application version as a tuple of integers.

For PEP 8 compliance, this specifier has the canonical name
``__version_info__`` rather than that of a typical global (e.g.,
``VERSION_PARTS``).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **simulation pipeline runner** (i.e., simulation activity
iteratively run by its parent pipeline) functionality.
'''

#FIXME: Refactor the "SimPipeRunnerMetadata.kind" instance variable from a
#non-type-safe string to a type-safe enumeration. Since the set of all types of
#runner methods supported by a given pipeline is dynamically defined by the
#current implementation of the subclass for that pipeline, we'll need to
#dynamically define one eneumeration type for each pipeline subclass at runtime
#in a presumably clever manner.

# ....................{ IMPORTS                           }....................
import functools
from betse.exceptions import (
    BetseSimPipeException, BetseSimPipeRunnerUnsatisfiedException)
from betse.science.phase.phasecls import SimPhase
from betse.science.phase.require.abc.phasereqset import (
    SimPhaseRequirements, SimPhaseRequirementsOrNoneTypes)
from betse.util.io.log import logs
from betse.util.type.text.string import strs
from betse.util.type.types import type_check, CallableTypes, SequenceTypes

# ....................{ CLASSES                           }....................
class SimPipeRunnerMetadata(object):
    '''
    **Simulation pipeline runner** (i.e., method of a :class:`SimPipeABC`
    subclass decorated by the :func:`piperunner` decorator) metadata.

    This metadata is available via the ``metadata`` instance variable of each
    such runner method.

    Attributes (Pipeline)
    ----------
    noun_singular_lowercase : str
        Human-readable lowercase singular noun synopsizing this type of runner
        (e.g., ``animation``, ``plot``), equivalent to the
        :attr:`SimPipeABC.subclass._noun_singular_lowercase` constant defined
        by the parent pipeline of this runner.
    noun_singular_uppercase : str
        Human-readable uppercase singular noun synopsizing this type of runner
        (e.g., ``animation``, ``plot``), equivalent to the
        :attr:`SimPipeABC.subclass._noun_singular_uppercase` constant defined
        by the parent pipeline of this runner.
    verb_continuous : str
        Human-readable verb in the continuous tense synopsizing the type of
        action performed by this type of runner (e.g., ``Saving``), equivalent
        to the :attr:`SimPipeABC.subclass._VERB_CONTINUOUS` constant defined by
        the parent pipeline of this runner.

    Attributes (Runner)
    ----------
    categories : SequenceTypes
        Sequence of one or more human-readable strings iteratively naming all
        arbitrary categories to which this runner belongs (in descending order
        of hierarchical taxonomy).
    kind : str
        Machine-readable type of this runner as a raw string. Although this
        string is currently equivalent to :attr:`method_name` excluding the
        :attr:`SimPipeABC._RUNNER_METHOD_NAME_PREFIX` substring required by the
        parent pipeline of this runner, callers should *not* assume this
        low-level implementation detail to always be.
    method_name : str
        Machine-readable name of the method implementing this runner.
    requirements : SimPhaseRequirements
        Immutable set of zero or more :class:`SimPhaseRequirement` instances
        specifying all simulation features required by this runner.
    description : str
        Human-readable description of this runner as a **single-line string**
        (i.e., containing no newlines).

    See Also
    ----------
    :func:`piperunner`
        Further details.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self,

        # Mandatory parameters.
        method: CallableTypes,
        categories: SequenceTypes,

        # Optional parameters.
        requirements: SimPhaseRequirementsOrNoneTypes = None,
    ) -> None:
        '''
        Initialize this metadata.

        Parameters
        ----------
        method: CallableTypes
            Unbound method (i.e., function) associated with this metadata.
        categories : SequenceTypes
            Sequence of one or more human-readable category names.
        requirements : SimPhaseRequirementsOrNoneTypes
            Immutable set of zero or more :class:`SimPhaseRequirement`
            instances *or* ``None``, in which case this parameter defaults to
            the empty set of such instances.

        Raises
        ----------
        BetseSimPipeException
            If this method has no docstring.
        '''

        # Initialize our superclass with the passed method.
        super().__init__()

        # Default all unpassed parameters to sane defaults.
        if requirements is None:
            requirements = SimPhaseRequirements()

        # Classify all passed parameters.
        self.requirements = requirements
        self.categories = categories
        self.method_name = method.__name__

        # Nullify all remaining instance variables, subsequently defined by the
        # "SimPipeABCMeta" metaclass at "SimPipeABC" subclass definition time
        # following the definition of this runner method.
        self.kind = None
        self.noun_singular_lowercase = None
        self.noun_singular_uppercase = None
        self.verb_continuous = None

        # Default this runner's description to its docstring.
        self.description = method.__doc__

        # If this docstring is empty, raise an exception.
        if not self.description:
            raise BetseSimPipeException(
                'Pipeline runner method {}() docstring undefined.'.format(
                    self.method_name))
        # Else, this docstring is non-empty.

        # Transform this docstring into a description by...
        self.description = (
            # Removing all leading and trailing whitespace.
            strs.remove_whitespace_presuffix(
            # Reducing from a (possibly) multi- to single-line string.
            strs.unwrap(self.description)))

        # Expose the metadata associated with this runner to callers. Due to
        # Python constraints on class decorators, *ONLY* the bound
        # SimPipeRunner.__call__() method returned by this method below is
        # accessible to callers. Notably, this "SimPipeRunner" instance is not
        # accessible to callers. Ergo, attaching this method to this method is
        # the only means of exposing this metadata to callers. *shrug*
        method.metadata = self

# ....................{ DECORATORS                        }....................
@type_check
def piperunner(
    # Mandatory metadata.
    categories: SequenceTypes,

    # Optional metadata.
    requirements: SimPhaseRequirementsOrNoneTypes = None,
) -> CallableTypes:
    '''
    Decorator annotating **simulation pipeline runners** (i.e.,
    :meth:`SimPipeRunner.__call__` subclasses with names prefixed by
    :attr:`SimPipeABC._RUNNER_METHOD_NAME_PREFIX`) with custom metadata.

    All methods decorated by this decorator are guaranteed to be instances of
    the :class:`SimPipeRunner` class, which provides all metadata passed to
    this decorator as instance variables of the same name.

    Caveats
    ----------
    **This decorator is strictly optional.** Runners *not* decorated by this
    decorator are still runnable from simulation pipelines. Since this
    decorator annotates runners with metadata, however, unannotated runners
    will *not* be usable by external interfaces expecting this metadata --
    typically, GUIs populating interactive widget fields by this metadata.

    **Runner methods decorated by this decorator should not be decorated by
    other decorators.** In particular, decorated methods should *not* also be
    decorated by :func:`@type_check`, which this decorator already internally
    decorates all decorated methods by.

    Parameters
    ----------
    categories : SequenceTypes
        Sequence of one or more human-readable strings iteratively naming all
        arbitrary categories to which this runner belongs (in descending order
        of hierarchical taxonomy). Categories are arbitrary labels accessed
        *only* by external interfaces and are otherwise ignored by the core
        codebase. Specifically:

        * The first string in this sequence names an arbitrary **root
          category** (e.g., root node in a tree view), intended to be shared
          between multiple runners. This string is typically a broadly
          applicable label such as ``Voltage Plots``.
        * The last string in this sequence names an arbitrary **leaf category**
          (e.g., leaf node in a tree view), intended to be unique to a single
          runner. This string is typically a narrowly applicable label such as
          ``Extracellular Voltage Plot``.
        * All other strings in this sequence name arbitrary categories of
          increasingly fine-grained depth, again intended to be shared between
          multiple runners.
    requirements : SimPhaseRequirementsOrNoneTypes
        Immutable set of zero or more :class:`SimPhaseRequirement` instances
        specifying all simulation features required by this runner. This
        decorator then decorates this runner by performing the following logic
        immediately *before* calling this runner:

        * For each requirement in this set...

          * If this requirement is unsatisfied by the current simulation phase
            (e.g., as the configuration for this phase disables extracellular
            spaces), this requirement (and hence this runner) is unsatisfied.
            Since this constitutes a fatal error, an
            :class:`BetseSimPipeRunnerUnsatisfiedException` is raised.
          * Else, this runner is run.

        Defaults to ``None``, in which case no such decoration is applied.
    '''

    @type_check
    def _piperunner_method_factory(method: CallableTypes) -> CallableTypes:
        '''
        Closure both type-checking *and* annotating the passed simulation
        pipeline runner method with the metadata passed to the outer decorator
        defining this closure, returning an instance of the class decorator
        exposing this metadata to external interfaces.

        Parameters
        ----------
        method : CallableTypes
            Unbound method (i.e., function) to be decorated by (in order):

            #. The :func:`@type_check` decorator, type checking this method.
               For efficiency, callers should ensure this method is *not*
               externally decorated by this decorator.
            #. The :class:`SimPipeRunner` class decorator,
               annotating this method with this metadata.

        See Also
        ----------
        :func:`piperunner`
            Further details.
        '''

        # As a caller convenience, ensure this method is type-checked.
        method_typed = type_check(method)

        @functools.wraps(method_typed)
        @type_check
        def _piperunner_method(
            # To avoid circular import dependencies, this is validated as a
            # fully-qualified class name resolved at runtime.
            #
            # For clarity, this parameter has been renamed from the customary
            # "self" nomenclature for a bound method.
            self_pipeline: 'betse.science.pipe.pipeabc.SimPipeABC',
            phase: SimPhase,
            *args, **kwargs
        ) -> object:
            '''
            Closure validating this simulation pipeline runner to be satisfied
            by the passed simulation pipeline and phase *before* running that
            runner by calling the method passed to the outer decorator defining
            this closure with the metadata passed to the same decorator,
            returning the values returned by that method call.

            Specifically, this closure either:

            * If this runner is **unsatisfied** (i.e., requires one or more
              simulation features disabled by the user-defined simulation
              configuration for this phase), raises an exception.
            * Else, logs an attempt to run this runner.

            Parameters
            ----------
            self_pipeline : SimPipeABC
                Current simulation pipeline.
            phase : SimPhase
                Current simulation phase.

            All remaining parameters are passed as is to the method passed to
            the outer decorator defining this closure.

            Raises
            ----------
            BetseSimVisualUnsatisfiedException
                If this runner is unsatisfied.

            See Also
            ----------
            :func:`piperunner`
                Further details.
            '''

            # Metadata associated with this runner.
            runner_metadata = _piperunner_method.metadata

            # If any runner requirement is unsatisfied, raise an exception.
            for requirement in runner_metadata.requirements:
                if not requirement.is_satisfied(phase):
                    raise BetseSimPipeRunnerUnsatisfiedException(
                        result='{} "{}" requirement unsatisfied'.format(
                            runner_metadata.noun_singular_uppercase,
                            runner_metadata.kind),
                        reason='{} disabled'.format(requirement.name),
                    )
            # Else, all runner requirements are satisfied.

            # Log the subsequent attempt to run this runner.
            logs.log_info(
                '%s %s "%s"...',
                runner_metadata.verb_continuous,
                runner_metadata.noun_singular_lowercase,
                runner_metadata.kind)

            # Else, this runner is satisfied. Since the prior call already
            # logged the attempt to run this runner, avoid redoing so here.
            #
            # Simply call this method to run this runner.
            return method_typed(self_pipeline, phase, *args, **kwargs)

        # Expose this metadata as an instance variable of this method closure.
        _piperunner_method.metadata = SimPipeRunnerMetadata(
            method=method_typed,
            categories=categories,
            requirements=requirements,
        )

        # Return this method closure.
        return _piperunner_method

    # Return this method factory closure.
    return _piperunner_method_factory



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all post-simulation plot subclasses.
'''

# ....................{ IMPORTS                            }....................
from betse.science.phase.phasecls import SimPhase
from betse.science.visual.visabc import VisualCellsABC
from betse.util.type.types import type_check

# ....................{ CLASSES                            }....................
class PlotCellsAfterSolving(VisualCellsABC):
    '''
    Abstract base class of all post-simulation cell plot subclasses, plotting
    simulation data over the cell cluster _after_ rather than _during_
    simulation modelling.
    '''

    @type_check
    def __init__(
        self,
        phase: SimPhase,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this post-simulation plot.

        Parameters
        ----------
        phase: SimPhase
            Current simulation phase.

        See the superclass `__init__()` method for all remaining parameters.
        '''

        # Initialize our superclass.
        super().__init__(
            *args,

            # Pass this simulation phase as is to our superclass.
            phase=phase,

            # Save and show this post-simulation plot only if this configuration
            # enables doing so.
            is_save=phase.p.plot.is_after_sim_save,
            is_show=phase.p.plot.is_after_sim_show,

            # Save all post-simulation plots to the same parent directory.
            save_dir_parent_basename='plot',

            # Pass all remaining arguments as is to our superclass.
            **kwargs
        )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **path archival** (i.e., compression and decompression of both
directories _and_ non-directory files) facilities.
'''

#FIXME: Add a new convert_format_to_filetype() function defined as follows
#
#    @type_check
#    def convert_format_to_filetype(format: str) -> StrOrNoneType:
#
#        #FIXME: Implement this format. Basically, we need to define an additional
#        #private sequence global -- say:
#        #
#        #    _ARCHIVE_FILETYPES_PREFERRED = ('bosc', 'lzo', 'xz', 'bz2', 'gz')
#        #
#        #Given that, we then need to perform the intersection of that sequence with
#        #the "ARCHIVE_FILETYPES" set, producing a new sequence containing only the
#        #elements of that sequence supported by the current system. Lastly, we then
#        #need to return the *FIRST* element of that sequence as the implementation
#        #of this format. For safety, note that this intersection could
#        #technically be empty, in which case None should be silently returned.
#        if format == 'auto':
#            ...  # something something
#        elif format == 'none':
#            return None
#        elif format in ARCHIVE_FILETYPES:
#            return format
#        #FIXME: It'd be great to raise a more comprehensive exception, but
#        #this would certainly do for the moment.
#        else:
#            raise BetseArchiveException(
#                'Archive format "{}" unsupported.'.format(format))
#FIXME: Refactor the "FILE HANDLING" section of our YAML file to resemble:
#
#path:
#  seed:
#    object:
#      file: SEEDS/world_1.betse  # File containing all cell cluster objects.
#
#  init:             # Paths for initialization saving and loading.
#    object:
#      file: INITS/init_1.betse   # File containing all initialization objects.
#    export:
#      directory: RESULTS/init_1  # Directory containing all initialization
#                                 # exports (e.g., plots, animations).
#
#  sim:
#    object:
#      file: SIMS/sim_1.betse    # filename to load/save sim run
#    export:
#      directory: RESULTS/sim_1  # directory to load/save data, plots, and animations for sim
#
#  object:
#    compression: auto  # Supported object compression formats include:
#                       # "auto" (smart compression), "none" (no compression),
#                       # "gz" (fast minimal compression),
#                       # "bz2" (medium compression), and
#                       # "xz" (slow maximal compression).
#FIXME: Pass the current value of the above "path/object/compression" key to the
#convert_format_to_filetype() function, producing a filetype suitable for use in
#pickling and unpickling compressed archives of that compression type.

#FIXME: Add support for additional non-stdlib archive packages, including:
#
#* https://github.com/Blosc/python-blosc, "...a Python wrapper for the extremely
#  fast Blosc compression library."
#* https://github.com/Iotic-Labs/py-lz4framed and
#  https://github.com/python-lz4/python-lz4, competing Python bindings for the
#  LZ4 compression library.
#* https://github.com/sergey-dryabzhinsky/python-zstd and
#  https://github.com/indygreg/python-zstandard, competing Python bindings for
#  the ZSTD compression library.
#
#While all of the above packages are released under BSD-compatible licenses,
#note that the following non-stdlib archive packages are strictly constrained by
#GPL licenses and *MUST* thus be avoided:
#
#* https://github.com/jd-boyd/python-lzo.
#
#Since the already supported XZ (LZMA) format arguably satisfies the ideal
#tradeoff between compression speed and ratio for our purposes, there may not
#exist much incentive to actually implement support for additional non-stdlib
#formats. Nonetheless, the future exists and it is always coming.

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseArchiveException
from betse.util.type.types import type_check
from io import BufferedIOBase

# ....................{ CONSTANTS ~ public                }....................
# Initialized below by the init() function.
ARCHIVE_FILETYPES = set()
'''
Set of all archive filetypes supported by this submodule.

Tradeoffs
----------
Most compression algorithms succumb to the typical tradeoff between time and
space complexity. Specifically, as the **(de)compression speed** (i.e.,
duration of time required to compress or decompress a given sequence of bytes)
of a compression algorithm increases, the **compression ratio** (i.e., filesize
of the output compressed archive to the size in bytes of the uncompressed
input) of that algorithm typically increases. The principal exception to this
heuristic is BZIP2, which runs slower and compresses less than competing
algorithms.

Most benchmarks qualitatively profile the algorithms supported by this
submodule as follows (in increasing order of compression ratio):

* GZIP, exhibiting the worst compression ratio and fastest compression speed.
* BZIP2, exhibiting average compression ratio and slowest compression speed.
* XZ (effectively equivalent to LZMA), exhibiting the best compression ratio
  and average compression speed.

For general-purpose use, the **XZ** algorithm arguably strikes ideal balance
between compression ratio and speed. Most algorithms exhibit similar
decompression speed, which is thus ignorable for comparison purposes.

Caveats
----------
This set contains neither the ``tar`` nor ``zip`` container filetypes, which
remain intentionally unsupported by this submodule. Neither the standard
:mod:`tarfile` module supporting ``tar`` files nor the standard :mod:`zipfile`
module supporting ``zip`` files provide a :class:`file`-like API for reading
and writing arbitrary bytes or characters to and from these files, the primary
use case and hence mandate of this submodule.

While wrapping the container-oriented APIs provided by these modules with
:class:`file`-like APIs is feasible, there exists no demonstrable incentive to
do so. All compression formats supported by these modules are conveniently
already supported by :class:`file`-like APIs implemented by other standard
modules (e.g., the :class:`bz2.BZ2File` and :class:`gzip.GzipFile` classes).
'''

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_unless_filetype(pathname: str) -> None:
    '''
    Raise an exception unless the passed pathname is suffixed by a filetype
    corresponding to a supported archive format.

    Parameters
    ----------
    pathname : str
        Relative or absolute path of the file to be validated.

    See Also
    ----------
    :func:`is_filetype`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames
    from betse.util.py.module import pymodname
    from betse.util.type.text.string import strjoin

    # If this pathname is *NOT* suffixed by a supported archive filetype...
    if not is_filetype(pathname):
        # Filetype of this pathname if any or raise an exception otherwise.
        filetype = pathnames.get_filetype_undotted(pathname)

        # Fully-qualified name of the optional stdlib module required to write
        # archives of this filetype if this is a recognized (but unsupported)
        # archive filetype or None otherwise.
        module_name = _ARCHIVE_FILETYPE_TO_MODULE_NAME.get(filetype, None)

        # If this is a recognized archive filetype, this module *CANNOT* have
        # been installed with Pythoni, implying the corresponding third-party
        # library headers to have been unavailable at Python installation time.
        if module_name is not None:
            # Exception message to be raised describing this constraint.
            exception_message = (
                'Module "{}" not found '
                '(typically due to third-party library headers '
                'required by this module being unavailable at '
                'Python installation time).'.format(module_name))

            # Raise the expected exception.
            pymodname.die_unless_module(module_name, exception_message)

            # The above call should have raised an exception. For safety,
            # unconditionally raise another exception as a sentinel.
            raise ValueError('That which should not be has become.')
        #  Else, this is an unrecognized archive filetype.
        else:
            # Human-readable string listing all supported archive filetypes.
            archive_filetypes = strjoin.join_as_disconjunction_double_quoted(
                *ARCHIVE_FILETYPES)

            # Raise an exception embedding this string.
            raise BetseArchiveException(
                'Filename "{}" not suffixed by a '
                'supported archive format (i.e., {}).'.format(
                    pathname, archive_filetypes))

# ....................{ TESTERS                           }....................
@type_check
def is_filetype(pathname: str) -> bool:
    '''
    ``True`` only if the passed pathname is suffixed by a filetype
    corresponding to a supported archive format.

    ``True`` is returned only if the **rightmost filetype** (i.e., substring
    following the last ``.`` character) of this pathname is in the
    :data:`ARCHIVE_FILETYPES` set global. All preceding filetypes of this
    pathname are ignored.

    Parameters
    ----------
    pathname : str
        Relative or absolute path of the file to be tested.

    Returns
    ----------
    bool
        `True` only if this pathname is suffixed by an archive filetype.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames

    # True only if this pathname is suffixed by an archive filetype.
    return pathnames.is_filetype_undotted_in(
        pathname=pathname, filetypes_undotted=ARCHIVE_FILETYPES)

# ....................{ IO                                }....................
@type_check
def read_bytes(filename: str) -> BufferedIOBase:
    '''
    Open and return a filehandle suitable for reading the binary archive file
    with the passed filename.

    This function returns a :class:`file`-like object suitable for use wherever
    the :func:`open` builtin is callable (e.g., in ``with`` statements).

    Parameters
    ----------
    filename : str
        Relative or absolute path of the binary archive file to be read, whose
        **rightmost filetype** (i.e., substring suffixing the last ``.``
        character in this pathname) *must* be that of a supported archive
        format.

    Returns
    ----------
    BufferedIOBase
        :class:`file`-like object encapsulating this opened file.

    Raises
    ----------
    BetseArchiveException
        If this filename is *not* suffixed by a supported archive filetype.
    BetseFileException
        If this filename is *not* that of an existing file.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import files, pathnames

    # Raise an exception unless this filename has a supported archive filetype
    # *AND* is an existing file.
    die_unless_filetype(filename)
    files.die_unless_file(filename)

    # Filetype of this pathname. By the above validation, this filetype is
    # guaranteed to both exist and be a supported archive filetype. Hence, no
    # additional validation is required.
    filetype = pathnames.get_filetype_undotted(filename)

    # Callable reading archives of this filetype.
    reader = _ARCHIVE_FILETYPE_TO_READER[filetype]

    # Open and return a filehandle suitable for reading this archive.
    return reader(filename)


@type_check
def write_bytes(filename: str, is_overwritable: bool = False) -> (
    BufferedIOBase):
    '''
    Open and return a filehandle suitable for writing the binary archive file
    with the passed filename.

    This function returns a :class:`file`-like object suitable for use wherever
    the :func:`open` builtin is callable (e.g., in ``with`` statements).

    Parameters
    ----------
    filename : str
        Relative or absolute path of the binary archive file to be written,
        whose **rightmost filetype** (i.e., substring suffixing the last ``.``
        character in this pathname) *must* be that of a supported archive
        format.
    is_overwritable : optional[bool]
        ``True`` if overwriting this file when this file already exists *or*
        ``False`` if raising an exception when this file already exists.
        Defaults to ``False`` for safety.

    Returns
    ----------
    BufferedIOBase
        :class:`file`-like object encapsulating this opened file.

    Raises
    ----------
    BetseArchiveException
        If this filename is *not* suffixed by a supported archive filetype.
    BetsePathException
        If this filename is that of an existing file or directory.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import dirs, paths, pathnames

    # Raise an exception unless this filename has a supported archive filetype.
    die_unless_filetype(filename)

    # If this file is *NOT* overwritable, raise an exception if this path
    # already exists.
    if not is_overwritable:
        paths.die_if_path(filename)

    # Create the parent directory of this file if needed.
    dirs.make_parent_unless_dir(filename)

    # Filetype of this pathname. By the above validation, this filetype is
    # guaranteed to both exist and be a supported archive filetype. Hence, no
    # additional validation is required.
    filetype = pathnames.get_filetype_undotted(filename)

    # Callable writing archives of this filetype.
    writer = _ARCHIVE_FILETYPE_TO_WRITER[filetype]

    # Open and return a filehandle suitable for writing this archive.
    return writer(filename, is_overwritable=is_overwritable)

# ....................{ IO ~ bz2                          }....................
def _read_bytes_bz2(filename: str) -> BufferedIOBase:
    '''
    Open and return a filehandle suitable for reading the binary bzip-archived
    file with the passed filename.

    This private function is intended to be called *only* by the public
    :func:`reading_bytes` function.
    '''

    # This optional stdlib module is guaranteed to exist and hence be safely
    # importable here, due to the above die_unless_filetype() call.
    from bz2 import BZ2File

    # Open and return a filehandle suitable for reading this file.
    return BZ2File(filename, mode='rb')


def _write_bytes_bz2(filename: str, is_overwritable: bool) -> BufferedIOBase:
    '''
    Open and return a filehandle suitable for writing the binary bzip-archived
    file with the passed filename.

    This private function is intended to be called *only* by the public
    :func:`writing_bytes` function.
    '''

    # Avoid circular import dependencies.
    from betse.util.io import iofiles

    # This optional stdlib module is guaranteed to exist and hence be safely
    # importable here, due to the above die_unless_filetype() call.
    from bz2 import BZ2File

    # Open and return a filehandle suitable for e(x)clusively writing this file.
    return BZ2File(
        filename, mode=iofiles.get_mode_write_bytes(is_overwritable))

# ....................{ IO ~ gz                           }....................
def _read_bytes_gz(filename: str) -> BufferedIOBase:
    '''
    Open and return a filehandle suitable for reading the binary gzip-archived
    file with the passed filename.

    This private function is intended to be called *only* by the public
    :func:`reading_bytes` function.
    '''

    # This optional stdlib module is guaranteed to exist and hence be safely
    # importable here, due to the above die_unless_filetype() call.
    from gzip import GzipFile

    # Open and return a filehandle suitable for reading this file.
    return GzipFile(filename, mode='rb')


def _write_bytes_gz(filename: str, is_overwritable: bool) -> BufferedIOBase:
    '''
    Open and return a filehandle suitable for writing the binary gzip-archived
    file with the passed filename.

    This private function is intended to be called *only* by the public
    :func:`writing_bytes` function.
    '''

    # Avoid circular import dependencies.
    from betse.util.io import iofiles

    # This optional stdlib module is guaranteed to exist and hence be safely
    # importable here, due to the above die_unless_filetype() call.
    from gzip import GzipFile

    # Open and return a filehandle suitable for e(x)clusively writing this file.
    return GzipFile(
        filename, mode=iofiles.get_mode_write_bytes(is_overwritable))

# ....................{ IO ~ xz                           }....................
def _read_bytes_xz(filename: str) -> BufferedIOBase:
    '''
    Open and return a filehandle suitable for reading the binary LZMA-archived
    file with the passed filename.

    This private function is intended to be called *only* by the public
    :func:`reading_bytes` function.
    '''

    # This optional stdlib module is guaranteed to exist and hence be safely
    # importable here, due to the above die_unless_filetype() call.
    from lzma import LZMAFile

    # Open and return a filehandle suitable for reading this file.
    return LZMAFile(filename, mode='rb')


def _write_bytes_xz(filename: str, is_overwritable: bool) -> BufferedIOBase:
    '''
    Open and return a filehandle suitable for writing the binary LZMA-archived
    file with the passed filename.

    This private function is intended to be called *only* by the public
    :func:`writing_bytes` function.
    '''

    # Avoid circular import dependencies.
    from betse.util.io import iofiles

    # This optional stdlib module is guaranteed to exist and hence be safely
    # importable here, due to the above die_unless_filetype() call.
    from lzma import LZMAFile

    # Open and return a filehandle suitable for e(x)clusively writing this file.
    return LZMAFile(
        filename, mode=iofiles.get_mode_write_bytes(is_overwritable))

# ....................{ CONSTANTS ~ private               }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: When adding a new key-value pair or removing an existing key-value
# pair from the dictionaries defined below, all testing parametrizations
# exercising these dictionaries (e.g., in the
# "betse_test.unit.path.test_archive" submodule) *MUST* be updated accordingly.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

_ARCHIVE_FILETYPE_TO_MODULE_NAME = {
    'bz2': 'bz2',
    'gz':  'gzip',
    'xz':  'lzma',
}
'''
Dictionary mapping from each archive filetype supported by this submodule to
the fully-qualified name of the optional stdlib module required to write and
read archives of this filetype.
'''


_ARCHIVE_FILETYPE_TO_READER = {
    'bz2': _read_bytes_bz2,
    'gz':  _read_bytes_gz,
    'xz':  _read_bytes_xz,
}
'''
Dictionary mapping from each archive filetype supported by this submodule to
the private function of this submodule reading archives of this filetype.
'''


_ARCHIVE_FILETYPE_TO_WRITER = {
    'bz2': _write_bytes_bz2,
    'gz':  _write_bytes_gz,
    'xz':  _write_bytes_xz,
}
'''
Dictionary mapping from each archive filetype supported by this submodule to
the private function of this submodule writing archives of this filetype.
'''

# ....................{ INITIALIZERS                      }....................
def _init() -> None:
    '''
    Initialize this submodule.

    This function defines all currently undefined global submodule variables.
    '''

    # Avoid circular import dependencies.
    from betse.util.py.module import pymodname

    # For each supported archive filetype and the fully-qualified name of the
    # optional stdlib module required to write archives of this filetype....
    for filetype, module_name in _ARCHIVE_FILETYPE_TO_MODULE_NAME.items():
        # If this module is installed...
        if pymodname.is_module(module_name):
            # Add this filetype to the set of all supported archive filetypes.
            ARCHIVE_FILETYPES.add(filetype)

# ....................{ MAIN                              }....................
# Initialize this submodule.
_init()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
YAML-backed simulation subconfigurations for exporting animations.
'''

# ....................{ IMPORTS                           }....................
from betse.lib.yaml.yamlalias import yaml_alias, yaml_alias_int_positive
from betse.lib.yaml.abc.yamlabc import YamlABC
from betse.lib.yaml.abc.yamllistabc import YamlList, YamlListItemABC
from betse.science.config.export.confexpabc import SimConfExportABC
from betse.science.config.export.visual.confexpvisabc import (
    SimConfVisualCellsYAMLMixin)
from betse.util.type.types import type_check, MappingType, SequenceTypes

# ....................{ SUBCLASSES                        }....................
class SimConfExportAnims(YamlABC):
    '''
    YAML-backed subconfiguration for exporting *all* animations (both
    in- and post-simulation) enabled by the current YAML-formatted simulation
    configuration file.

    Attributes (General)
    ----------
    is_overlay_current : bool
        ``True`` only if this configuration overlays streamlines (e.g.,
        electric current, concentration flux) onto appropriate animations.

    Attributes (While Solving)
    ----------
    is_while_sim : bool
        ``True`` only if this configuration displays and/or saves in-simulation
        animations.
    is_while_sim_save : bool
        ``True`` only if this configuration saves in-simulation animations.
    is_while_sim_show : bool
        ``True`` only if this configuration displays in-simulation animations.
    anim_while_sim : SimConfExportAnimCellsEmbedded
        Generic configuration applicable to all in-simulation animations.
        Ignored if :attr:``is_while_sim`` is ``False``.

    Attributes (After Solving)
    ----------
    is_after_sim : bool
        ``True`` only if this configuration displays and/or saves
        post-simulation animations.
    is_after_sim_save : bool
        ``True`` only if this configuration saves post-simulation animations.
    is_after_sim_show : bool
        ``True`` only if this configuration displays post-simulation
        animations.
    anims_after_sim : YamlList
        YAML-backed list of all post-simulation animations to be animated.
        Ignored if :attr:`is_after_sim` is ``False``.

    Attributes (Images)
    ----------
    is_images_save : bool
        ``True`` only if this configuration saves animation frames as images.
    image_filetype : str
        Filetype of all image files saved by this configuration. Ignored if
        :attr:`is_images_save` is ``False``.
    image_dpi : int
        Dots per inch (DPI) of all image files saved by this configuration.
        Ignored if :attr:`is_images_save` is ``False``.

    Attributes (Video)
    ----------
    is_video_save : bool
        ``True`` only if this configuration saves animation frames as video.
    video_bitrate : int
        Bitrate in bits per second of all video files saved by this
        configuration. Ignored if :attr:`is_video_save` is ``False``.
    video_codec_names : SequenceTypes
        List of the names of all encoder-specific codecs with which to encode
        animations (in descending order of preference), automatically:

        * Selecting the first codec supported by the selected writer.
        * Replacing any codec named `auto` by the name of a recommended codec
          specific to the selected writer and filetype. For details, see the
          `betse.lib.matplotlib.writer.mplvideo.get_first_codec_name()`
          function.

        Ignored if :attr:`is_video_save` is ``False``.
    video_dpi : int
        Dots per inch (DPI) of all frames of all video files saved by this
        configuration. Ignored if :attr:`is_images_save` is ``False``.
    video_filetype : str
        Filetype of all video files saved by this configuration. Supported
        filetypes include:

        * ``mkv`` (Matroska), an open-standard audio and video container
          supporting all relevant codecs and hence the default.
        * ``avi``, Microsoft's obsolete proprietary audio and video container.
        * ``gif``, a proprietary image format supporting video animations.
        * ``ogv`` (Theora), Xiph's open-standard audio and video container.
        * ``mov`` (QuickTime), Apple's proprietary audio and video container.
        * ``mp4`` (MPEG-4 Part 14), an open-standard audio and video container.
        * ``webm`` (WebM), Google's proprietary audio and video container.

        Ignored if :attr:`is_video_save` is ``False``.
    video_framerate : int
        Framerate in frames per second of all video files saved by this
        configuration. Ignored if :attr:`is_video_save` is ``False``.
    video_metadata : MappingType
        Dictionary mapping from the alphabetic lowercase name of video metadata
        supported by the active video encoder to that metadata's human-readable
        string to be embedded in all video files saved by this configuration.
        Supported names include: ``title``, ``artist``, ``genre``, ``subject``,
        ``copyright``, ``srcform``, and ``comment``. If this dictionary does
        *not* contain a ``copyright`` key, this key will be automatically
        synthesized from the current year. Ignored if :attr:`is_video_save` is
        ``False``.
    video_writer_names : SequenceTypes
        List of the names of all matplotlib animation writers with which to
        encode animations (in order of descending preference), automatically
        selecting the first writer installed on the current system. Supported
        names include:

        * ``ffmpeg``, an open-source cross-platform audio and video encoder.
        * ``avconv``, an open-source cross-platform audio and video encoder
          forked from (and largely interchangeable with) ``ffmpeg``.
        * ``mencoder``, an open-source cross-platform audio and video encoder
          associated with MPlayer, a popular media player.
        * ``imagemagick``, an open-source cross-platform image manipulation
          suite supporting *only* creation of animated GIFs.

        Ignored if :attr:`is_video_save` is ``False``.
    '''

    # ..................{ ALIASES                           }..................
    is_overlay_current = yaml_alias(
        "['results options']['overlay currents']", bool)

    # ..................{ ALIASES ~ while                   }..................
    is_while_sim_save = yaml_alias(
        "['results options']['while solving']['animations']['save']", bool)
    is_while_sim_show = yaml_alias(
        "['results options']['while solving']['animations']['show']", bool)

    # ..................{ ALIASES ~ after                   }..................
    is_after_sim_save = yaml_alias(
        "['results options']['after solving']['animations']['save']", bool)
    is_after_sim_show = yaml_alias(
        "['results options']['after solving']['animations']['show']", bool)

    # ..................{ ALIASES ~ save : images           }..................
    is_images_save = yaml_alias(
        "['results options']['save']['animations']['images']['enabled']", bool)
    image_filetype = yaml_alias(
        "['results options']['save']['animations']['images']['filetype']", str)
    image_dpi = yaml_alias_int_positive(
        "['results options']['save']['animations']['images']['dpi']")

    # ..................{ ALIASES ~ save : video            }..................
    is_video_save = yaml_alias(
        "['results options']['save']['animations']['video']['enabled']", bool)
    video_bitrate = yaml_alias_int_positive(
        "['results options']['save']['animations']['video']['bitrate']")
    video_dpi = yaml_alias_int_positive(
        "['results options']['save']['animations']['video']['dpi']")
    video_filetype = yaml_alias(
        "['results options']['save']['animations']['video']['filetype']", str)
    video_framerate = yaml_alias_int_positive(
        "['results options']['save']['animations']['video']['framerate']")
    video_metadata = yaml_alias(
        "['results options']['save']['animations']['video']['metadata']",
        MappingType)
    video_writer_names = yaml_alias(
        "['results options']['save']['animations']['video']['writers']",
        SequenceTypes)
    video_codec_names = yaml_alias(
        "['results options']['save']['animations']['video']['codecs']",
        SequenceTypes)

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Encapsulate low-level dictionaries with high-level wrappers.
        self.anim_while_sim = SimConfExportAnimCellsEmbedded()

        # Encapsulate low-level lists of dictionaries with high-level wrappers.
        self.anims_after_sim = SimConfExportAnimCells.make_list()

    # ..................{ LOADERS                           }..................
    #FIXME: Default the "copyright" entry of video metadata to
    #"@ {}".format(current_year)".
    def load(self, *args, **kwargs) -> None:

        # Load our superclass with all passed arguments.
        super().load(*args, **kwargs)

        # Load all subconfigurations of this configuration.
        self.anim_while_sim.load(conf=self._conf[
            'results options']['while solving']['animations'])
        self.anims_after_sim.load(conf=self._conf[
            'results options']['after solving']['animations']['pipeline'])


    def unload(self) -> None:

        # Unload our superclass.
        super().unload()

        # Unload all subconfigurations of this configuration.
        self.anim_while_sim.unload()
        self.anims_after_sim.unload()

    # ..................{ PROPERTIES ~ while                }..................
    @property
    def is_while_sim(self) -> bool:
        return self.is_while_sim_save or self.is_while_sim_show


    @is_while_sim.setter
    @type_check
    def is_while_sim(self, is_while_sim: bool) -> None:
        self.is_while_sim_save = is_while_sim
        self.is_while_sim_show = is_while_sim

    # ..................{ PROPERTIES ~ after                }..................
    @property
    def is_after_sim(self) -> bool:
        return self.is_after_sim_save or self.is_after_sim_show


    @is_after_sim.setter
    @type_check
    def is_after_sim(self, is_after_sim: bool) -> None:
        self.is_after_sim_save = is_after_sim
        self.is_after_sim_show = is_after_sim

# ....................{ SUBCLASSES ~ item                 }....................
class SimConfExportAnimCells(
    SimConfVisualCellsYAMLMixin, SimConfExportABC):
    '''
    **Exported cell cluster animation subconfiguration** (i.e., YAML-backed
    list item configuring the exportation of one or more video files applicable
    to all cells from the simulation configuration file containing this item).
    '''

    # ..................{ SUPERCLASS                        }..................
    @classmethod
    @type_check
    def make_default(cls, yaml_list: YamlList) -> YamlListItemABC:

        # Duplicate the first animation in our default configuration file.
        return cls._make_loaded(conf={
            'name': yaml_list.get_item_name_uniquified(
                'Cell Cluster Plot ({})'),
            'type': 'voltage_membrane',
            'enabled': True,
            'colorbar': {
                'autoscale': True,
                'minimum': -70.0,
                'maximum':  10.0,
            },
        })


class SimConfExportAnimCellsEmbedded(SimConfVisualCellsYAMLMixin, YamlABC):
    '''
    **Exported embedded cell cluster animation subconfiguration** (i.e.,
    YAML-backed dictionary configuring the exportation of one or more video
    files applicable to all cells from the simulation configuration file
    containing this item).

    This subconfiguration unconditionally enables only a single animation and
    thus contains no distinguishing ``name``, ``type``, or ``enabled`` keys.
    '''

    pass



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **regex** (i.e., Python-compatible regular expression) facilities.
'''

# ....................{ IMPORTS                            }....................
import re
from betse.exceptions import BetseRegexException
from betse.util.type.types import (
    type_check,
    CallableOrStrTypes,
    IterableTypes,
    MappingType,
    RegexCompiledType,
    RegexMatchType,
    RegexMatchOrNoneTypes,
    RegexTypes,
    SequenceTypes,
    SequenceOrNoneTypes,
    StrOrNoneTypes,
)

# ....................{ FLAGS                              }....................
FLAG_MULTILINE = re.MULTILINE
'''
When specified, the pattern character:

* ``^`` matches both at the beginning of the subject string *and* at the
  beginning of each line (immediately following each newline) of this string.
* ``$`` matches both at the end of the subject string *and* at the end of each
  line (immediately preceding each newline) of this string.

By default:

* ``^`` matches only at the beginning of the subject string.
* ``$`` matches only at the end of the subject string and immediately before
  the newline (if any) at the end of this string.
'''

# ....................{ EXCEPTIONS                         }....................
@type_check
def die_unless_match(text: str, regex: RegexTypes, **kwargs) -> str:
    '''
    Raise an exception unless the passed subject string matches the passed
    regular expression.

    Raises
    ----------
    BetseRegexException
        If this subject string fails to match this regular expression.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strs

    # If this subject string fails to match this regex, raise an exception.
    if not is_match(text=text, regex=regex, **kwargs):
        # This subject string truncated to either the first newline in this
        # string *OR* the standard UNIX line length - whichever comes first.
        text_truncated = strs.truncate(
            text=text, suffix_prefix='\n', max_len=80)

        # Raise this exception with this truncated string for readability.
        raise BetseRegexException(
            'Subject string "{}" not matched by '
            'regular expression "{}".'.format(text_truncated, regex))


#FIXME: Rename to replace_substrs_line_or_die() for clarity.
@type_check
def die_unless_replace_substrs_line(
    text: str,
    regex: RegexTypes,
    replacement: CallableOrStrTypes,
    **kwargs
) -> str:
    '''
    Passed subject string with all substrings matching the passed regular
    expression replaced by the passed substitution in a line-oriented manner if
    this subject string contains at least one such substring *or* raise an
    exception otherwise.

    Raises
    ----------
    BetseRegexException
        If this subject string fails to match this regular expression.

    See Also
    ----------
    :func:`replace_substrs_line`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strs

    #FIXME: Generalize this validation into a new die_unless_match_line()
    #function of this submodule.

    # If this subject string fails to match this regex, raise an exception.
    #
    # Sadly, the re.sub() function called by the replace_substrs_line()
    # function called below fails to return metadata detailing the number of
    # replacements performed or even whether any replacements were performed.
    # Hence, this validation *MUST* be performed manually beforehand.
    if not is_match_line(text=text, regex=regex, **kwargs):
        # This subject string truncated to either the first newline in this
        # string *OR* the standard UNIX line length - whichever comes first.
        text_truncated = strs.truncate(
            text=text, suffix_prefix='\n', max_len=80)

        # Raise this exception with this truncated string for readability.
        raise BetseRegexException(
            'Subject string "{}" not matched by '
            'regular expression "{}".'.format(text_truncated, regex))

    # Else, this subject string matches this regex. In this case, perform this
    # search-and-replacement and return the result.
    return replace_substrs_line(
        text=text,
        regex=regex,
        replacement=replacement,
        **kwargs
    )

# ....................{ TESTERS                            }....................
def is_match(text: str, regex: RegexTypes, **kwargs) -> bool:
    '''
    ``True`` only if zero or more characters anchored to the beginning of the
    passed string match the passed regular expression.

    Parameters
    ----------
    text : str
        String to match.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as the
    :func:`re.match` function.

    Returns
    ----------
    bool
        ``True`` only if this string matches this regular expression.
    '''

    return get_match_or_none(text, regex, **kwargs) is not None


def is_match_line(text: str, regex: RegexTypes, **kwargs) -> bool:
    '''
    ``True`` only if one or more lines of the passed subject string match the
    passed regular expression.

    Parameters
    ----------
    text : str
        Subject string to match.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as the
    :func:`re.search` function.

    Returns
    ----------
    bool
        ``True`` only if this string matches this regular expression.

    See Also
    ----------
    :func:`get_match_line_first_or_none`
        Further details.
    '''

    return get_match_line_first_or_none(text, regex, **kwargs) is not None

# ....................{ MATCHERS ~ group : named           }....................
def get_match_groups_named(
    text: str, regex: RegexTypes, **kwargs) -> MappingType:
    '''
    Dictionary mapping explicitly named groups to substrings matched anchored
    to the beginning of the passed string against the passed regular expression
    if a match exists or raise an exception otherwise.

    For each capture group of the form ``(?P<group_name>...)`` in this regular
    expression, this dictionary contains a key-value pair whose:

    * Key is this named group's ``group_name``.
    * Value is either:

      * ``None`` if this group is unmatched by this string.
      * The matched substring otherwise.

    Unnamed (i.e., only numbered) groups are ignored and hence excluded from
    this dictionary, regardless of whether any of these groups matched. If
    undesirable, call the :func:`get_match_groups_numbered` function instead.

    Caveats
    ----------
    Python does *not* support the increasingly standardized named group form
    ``(?<group_name>...)`` -- only the Python-specific form
    ``(?P<group_name>...)``.

    Parameters
    ----------
    text : str
        String to match.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as the
    :func:`re.match` function.

    Returns
    ----------
    MappingType
        Dictionary mapping matched named groups.

    Raises
    ----------
    BetseRegexException
        If this string does *not* match this expression.

    See Also
    ----------
    :func:`get_match_or_none`
        Further details on regular expressions and keyword arguments.
    '''

    return get_match(text, regex, **kwargs).groupdict()

# ....................{ MATCHERS ~ group : number          }....................
def get_match_groups_numbered(
    text: str, regex: RegexTypes, **kwargs) -> SequenceTypes:
    '''
    Sequence of all groups matched anchored to the beginning of the passed
    subject string against the passed regular expression (ordered by the
    left-to-right lexical position at which each such group was matched) if any
    *or* raise an exception otherwise.

    Unmatched groups will have the value `None`.

    Parameters
    ----------
    text : str
        Subject string to match.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as the
    :func:`re.match` function.

    Returns
    ----------
    SequenceTypes
        Sequence of all matched groups.

    Raises
    ----------
    BetseRegexException
        If this string does *not* match this expression.

    See Also
    ----------
    :func:`get_match_or_none`
        Further details.
    '''

    return get_match(text, regex, **kwargs).groups()


def get_match_groups_numbered_or_none(
    text: str, regex: RegexTypes, **kwargs) -> SequenceOrNoneTypes:
    '''
    Sequence of all groups matched anchored to the beginning of the passed
    subject string against the passed regular expression (ordered by the
    left-to-right lexical position at which each such group was matched) if any
    *or* ``None`` otherwise.

    Unmatched groups will have the value ``None``.

    Parameters
    ----------
    text : str
        Subject string to match.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as the
    :func:`re.match` function.

    Returns
    ----------
    SequenceOrNoneTypes
        Either:
        * If this string matches this regular expression, the sequence of all
          groups matched from this string.
        * Else, ``None``.

    See Also
    ----------
    :func:`get_match_or_none`
        Further details.
    '''

    match = get_match_or_none(text, regex, **kwargs)
    return match.groups() if match is not None else None

# ....................{ MATCHERS ~ group : number : index  }....................
@type_check
def get_match_group_first(
    text: str, regex: RegexTypes, **kwargs) -> str:
    '''
    First grouped substring anchored to the beginning of the passed subject
    string matched against the passed regular expression if any *or* raise an
    exception otherwise.

    Parameters
    ----------
    text : str
        String to match.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as the
    :func:`re.match` function.

    Returns
    ----------
    str
        First grouped substring anchored to the beginning of this string.

    Raises
    ----------
    BetseRegexException
        If either:

        * This string does *not* match this expression.
        * This expression defines *no* match groups.
        * No match groups defined by this expression matched this string.

    See Also
    ----------
    :func:`get_match_or_none`
        Further details on regular expressions and keyword arguments.
    '''

    # First grouped substring if any *OR* "None" otherwise.
    match_group_first = get_match_group_first_or_none(text, regex, **kwargs)

    # If no grouped substring is matched, raise an exception.
    if match_group_first is None:
        die_unless_match(text, regex, **kwargs)

    # Return this grouped substring.
    return match_group_first


@type_check
def get_match_group_first_or_none(
    text: str, regex: RegexTypes, **kwargs) -> StrOrNoneTypes:
    '''
    First grouped substring anchored to the beginning of the passed subject
    string matched against the passed regular expression if any *or* ``None``
    otherwise.

    Parameters
    ----------
    text : str
        String to match.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as the
    :func:`re.match` function.

    Returns
    ----------
    StrOrNoneTypes
        Either:

        * If this subject string matches this regular expression, the first
          grouped substring anchored to the beginning of this string.
        * Else, ``None``.

    Raises
    ----------
    BetseRegexException
        If this string does *not* match this expression.

    See Also
    ----------
    :func:`get_match_or_none`
        Further details on regular expressions and keyword arguments.
    '''

    match = get_match_or_none(text, regex, **kwargs)
    return match.group(1) if match is not None else None

# ....................{ MATCHERS ~ full : first            }....................
def get_match_full_first_or_none(
    text: str, regex: RegexTypes, **kwargs) -> SequenceOrNoneTypes:
    '''
    First complete substring matched anchored to the beginning of the passed
    subject string against the passed regular expression if any *or* ``None``
    otherwise.

    Parameters
    ----------
    text : str
        Subject string to match.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as the
    :func:`re.match` function.

    Returns
    ----------
    SequenceOrNoneTypes
        Either:

        * If this string matches this regular expression, the first complete
          substring matched from this string.
        * Else, ``None``.

    See Also
    ----------
    :func:`get_match_or_none`
        Further details.
    '''

    match = get_match_or_none(text, regex, **kwargs)
    return match.group(0) if match is not None else None

# ....................{ MATCHERS ~ obj                     }....................
def get_match(text: str, regex: RegexTypes, **kwargs) -> RegexMatchType:
    '''
    Match object obtained by matching zero or more characters anchored to the
    beginning of the passed subject string against the passed regular
    expression if any match exists *or* raise an exception otherwise.

    Parameters
    ----------
    text : str
        Subject string to match.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as the
    :func:`re.match` function.

    Returns
    ----------
    RegexMatchType
        This match object.

    Raises
    ----------
    BetseRegexException
        If this string does *not* match this expression.

    See Also
    ----------
    :func:`get_match_or_none`
        Further details on calling conventions.
    '''

    # Object matching this string against this expression.
    match = get_match_or_none(text, regex, **kwargs)

    # If no match was found, convert the non-fatal "None" returned by the
    # re.match() function into a fatal exception. By design, no callables in
    # the "re" module raise exceptions.
    if match is None:
        raise BetseRegexException(
            'Subject string "{}" not matched by '
            'regular expression "{}".'.format(text, regex))

    # Return this match.
    return match


@type_check
def get_match_or_none(
    text: str, regex: RegexTypes, **kwargs) -> RegexMatchOrNoneTypes:
    '''
    Match object obtained by matching zero or more characters anchored to the
    beginning of the passed subject string against the passed regular
    expression if any match exists *or* ``None`` otherwise.

    Parameters
    ----------
    text : str
        Subject string to match.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as
    :func:`re.match`.

    Match Flags
    ----------
    For convenience, the following match flags will be enabled by default:

    * :data:`re.DOTALL`, forcing the ``.`` special character to match any
      character including newline. By default, this character matches any
      character excluding newline. The former is almost always preferable.

    Returns
    ----------
    RegexMatchOrNoneTypes
        This match object if a match exists *or* ``None`` otherwise.

    See Also
    ----------
    https://docs.python.org/3/library/re.html#re.match
        Further details on regular expressions and keyword arguments.
    '''

    # Sanitize the passed match flags.
    _init_kwargs_flags_nonline(regex, kwargs)

    # Return the only object matching this string against this expression.
    return re.match(regex, text, **kwargs)

# ....................{ MATCHERS ~ obj : first             }....................
@type_check
def get_match_first_or_none(
    text: str, regex: RegexTypes, **kwargs) -> RegexMatchOrNoneTypes:
    '''
    First match object obtained by matching the passed subject string against
    the passed regular expression if any match exists *or* ``None`` otherwise.

    Parameters
    ----------
    text : str
        Subject string to match.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as
    :func:`re.match`.

    Match Flags
    ----------
    For convenience, the following match flags will be enabled by default:

    * :data:`re.DOTALL`, forcing the ``.`` special character to match any
      character including newline. By default, this character matches any
      character excluding newline. The former is almost always preferable.

    Returns
    ----------
    RegexMatchOrNoneTypes
        This match object if a match exists *or* ``None`` otherwise.

    See Also
    ----------
    https://docs.python.org/3/library/re.html#re.match
        Further details on regular expressions and keyword arguments.
    '''

    # Sanitize the passed match flags.
    _init_kwargs_flags_nonline(regex, kwargs)

    # Return the first object matching this string against this expression.
    return re.search(regex, text, **kwargs)

# ....................{ MATCHERS ~ obj : line : first      }....................
@type_check
def get_match_line_first_or_none(
    text: str, regex: RegexTypes, **kwargs) -> RegexMatchOrNoneTypes:
    '''
    First match object obtained by matching the passed subject string against
    the passed regular expression in a line-oriented manner if any such match
    exists *or* ``None`` otherwise.

    To ensure that only single lines are matched, this regular expression
    should typically be prefixed by the ``^`` special character and/or suffixed
    by the ``$`` special character, anchoring matches to the start and/or end
    of this subject string as a whole and each line of this string.

    Parameters
    ----------
    text : str
        Subject string to match.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as the
    :func:`re.search` function.

    Match Flags
    ----------
    For convenience, the following match flags will be enabled by default:

    * :data:`re.MULTILINE`, forcing the ``^`` and ``$`` special characters to
      match both at the start and end of this string *and* at the start and end
      of each line of this string. By default, these characters match only at
      the former. Line-oriented matching requires both, however.

    Returns
    ----------
    RegexMatchOrNoneTypes
        This match object if a match exists *or* ``None`` otherwise.

    See Also
    ----------
    https://docs.python.org/3/library/re.html#re.search
        Further details on regular expressions and keyword arguments.
    '''

    # Sanitize the passed match flags for line-oriented matching.
    _init_kwargs_flags_line(regex, kwargs)

    # Return the first object matching this string against this expression.
    return re.search(regex, text, **kwargs)

# ....................{ ITERATORS                          }....................
@type_check
def iter_matches(text: str, regex: RegexTypes, **kwargs) -> IterableTypes:
    '''
    Generator iteratively yielding each non-overlapping match at any position
    of the passed string against the passed regular expression as a match
    object.

    If no such match exists, this function successfully returns the empty
    generator rather than raising a fatal exception.

    Parameters
    ----------
    text : str
        Subject string to match on.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as the
    :func:`re.finditer` function.

    Match Flags
    ----------
    For convenience, the following match flags will be enabled by default:

    * :data:`re.DOTALL`, forcing the ``.`` special character to match any
      character including newline. By default, this character matches any
      character excluding newline. The former is almost always preferable.

    Returns
    ----------
    IterableTypes
        Generator yielding match objects (i.e., instances of `re.SRE_Match`).

    See Also
    ----------
    https://docs.python.org/3/library/re.html#re.search
        Further details on regular expressions and keyword arguments.
    '''

    # Sanitize the passed match flags.
    _init_kwargs_flags_nonline(regex, kwargs)

    # Return this generator.
    return re.finditer(regex, text, **kwargs)


@type_check
def iter_matches_line(text: str, regex: RegexTypes, **kwargs) -> IterableTypes:
    '''
    Generator iteratively yielding each non-overlapping match at any position
    of the passed string against the passed regular expression in a
    line-oriented manner as a match object.

    If no such match exists, this function successfully returns the empty
    generator rather than raising a fatal exception.

    To ensure that only single lines are matched, this regular expression
    should typically be prefixed by the `^` special character and/or suffixed
    by the `$` special character, thus anchoring matches to the start and/or
    end of this string as a whole and each line of this string.

    Match Flags
    ----------
    For convenience, the following match flags will be enabled by default:

    * :data:`re.MULTILINE`, forcing the ``^`` and ``$`` special characters to
      match both at the start and end of this string *and* at the start and end
      of each line of this string. By default, these characters match only at
      the former. Line-oriented matching requires both, however.

    Parameters
    ----------
    text : str
        Subject string to match on.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as the
    :func:`re.finditer` function.

    Returns
    ----------
    IterableTypes
        Generator yielding match objects (i.e., instances of
        :class:`re.SRE_Match`).
    '''

    # Sanitize the passed match flags for line-oriented matching.
    _init_kwargs_flags_line(regex, kwargs)

    # Return this generator.
    return re.finditer(regex, text, **kwargs)

# ....................{ REMOVERS                           }....................
def remove_substrs(text: str, regex: RegexTypes, **kwargs) -> str:
    '''
    Remove all substrings in the passed string matching the passed regular
    expression.

    Parameters
    ----------
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.

    This function accepts the same optional keyword arguments as
    :func:`re.sub`.

    Returns
    ----------
    str
        Passed string containing no substrings matching such regular expression.

    See Also
    ----------
    https://docs.python.org/3/library/re.html#re.sub
        Further details on regular expressions and keyword arguments.
    '''

    return replace_substrs(text=text, regex=regex, replacement='', **kwargs)

# ....................{ REPLACERS                          }....................
@type_check
def replace_substrs(
    text: str,
    regex: RegexTypes,
    replacement: CallableOrStrTypes,
    **kwargs
) -> str:
    '''
    Passed subject string with all substrings matching the passed regular
    expression replaced by the passed substitution in a non-line-oriented
    manner.

    Match Flags
    ----------
    For convenience, the following match flags will be enabled by default:

    * :data:`re.DOTALL`, forcing the ``.`` special character to match any
      character including newline. By default, this character matches any
      character excluding newline. The former is almost always preferable.

    Parameters
    ----------
    text : str
        Subject string to replace these substrings of.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.
    replacement : CallableOrStrTypes
        Substitution to be performed, either a:

        * String.
        * Callable (e.g., function, lambda, method).

    This function accepts the same optional keyword arguments as the
    :func:`re.sub` function.

    Returns
    ----------
    str
        Passed string with all substrings matching this regular expression
        globally replaced with this substitution.

    See Also
    ----------
    :func:`betse.util.type.text.strs.replace_substrs`
        Equivalent literal-based replacer.
    https://docs.python.org/3/library/re.html#re.sub
        Further details on regular expressions and keyword arguments.
    '''

    # Sanitize the passed match flags.
    _init_kwargs_flags_nonline(regex, kwargs)

    # Substitute, if you please.
    return re.sub(regex, replacement, text, **kwargs)

# ....................{ REPLACERS : line                   }....................
@type_check
def replace_substrs_line(
    text: str,
    regex: RegexTypes,
    replacement: CallableOrStrTypes,
    **kwargs
) -> str:
    '''
    Passed subject string with all substrings matching the passed regular
    expression in a line-oriented manner (if any) replaced by the passed
    substitution *or* this string as is otherwise.

    If this subject string contains no such substrings, this function silently
    reduces to a noop. If undesirable, consider calling the stricter
    :func:`die_unless_replace_substrs_line` function instead.

    Match Flags
    ----------
    For convenience, the following match flags will be enabled by default:

    * :data:`re.MULTILINE`, forcing the ``^`` and ``$`` special characters to
      match both at the start and end of this string *and* at the start and end
      of each line of this string. By default, these characters match only at
      the former. Line-oriented matching requires both, however.

    Parameters
    ----------
    text : str
        Subject string to replace these substrings of.
    regex : RegexTypes
        Regular expression to be matched. This object should be either of type:

        * :class:`str`, signifying an uncompiled regular expression.
        * :class:`Pattern`, signifying a compiled regular expression object.
    replacement : CallableOrStrTypes
        Substitution to be performed, either a:

        * String.
        * Callable (e.g., function, lambda, method).

    This function accepts the same optional keyword arguments as the
    :func:`re.sub` function.

    Returns
    ----------
    str
        Passed string with all substrings matching this regular expression
        globally replaced with this substitution.

    See Also
    ----------
    https://docs.python.org/3/library/re.html#re.sub
        Further details on regular expressions and keyword arguments.
    '''

    # Sanitize the passed match flags in a line-oriented manner.
    _init_kwargs_flags_line(regex, kwargs)

    # Substitute, if you please.
    return re.sub(regex, replacement, text, **kwargs)

# ....................{ COMPILERS                          }....................
@type_check
def compile_regex(regex: str, **kwargs) -> RegexCompiledType:
    '''
    Compile the passed uncompiled regular expression.

    Parameters
    ----------
    regex : str
        Uncompiled regular expression to be compiled.

    All remaining keyword parameters are passed as is to the :func:`re.compile`
    function.

    Returns
    -------
    RegexCompiledType
        This compiled regular expression.
    '''

    # Return this regular expression compiled.
    return re.compile(regex, **kwargs)

# ....................{ SUBSTITUTERS                       }....................
@type_check
def _init_kwargs_flags_nonline(regex: RegexTypes, kwargs: MappingType) -> None:
    '''
    Sanitize the list of match flags in the passed dictionary for
    non-line-oriented matching.

    Specifically, this function adds the :data:`re.DOTALL` flag to the integer
    value of the ``flags`` key of this dictionary (defaulting to zero if
    currently unset).
    '''

    # If this regular expression is already compiled, reduce to a noop. Why?
    # Because flags *CANNOT* be respecified after the compilation phase.
    if isinstance(regex, RegexCompiledType):
        return

    # Else, this regular expression is uncompiled. In this case, these flags
    # are safely modifiable as required.
    kwargs['flags'] = kwargs.get('flags', 0) | re.DOTALL


@type_check
def _init_kwargs_flags_line(regex: RegexTypes, kwargs: MappingType) -> None:
    '''
    Sanitize the list of match flags in the passed dictionary for line-oriented
    matching.

    Specifically, this function adds the :data:`re.MULTILINE` flag to the
    integer value of the ``flags`` key of this dictionary (defaulting to zero
    if currently unset).
    '''

    # If this regular expression is already compiled, reduce to a noop. Why?
    # Because flags *CANNOT* be respecified after the compilation phase.
    if isinstance(regex, RegexCompiledType):
        return

    # Else, this regular expression is uncompiled. In this case, these flags are
    # safely modifiable as required.
    kwargs['flags'] = kwargs.get('flags', 0) | re.MULTILINE
